import 'package:contentsquare/src/modules/screengraph/screengraph_node.dart';
import 'package:contentsquare/src/modules/screengraph/screengraph_utils.dart';

import 'package:flutter/widgets.dart';

class _ScreengraphNodeOrChildren {
  final Map? node;
  final List? children;

  _ScreengraphNodeOrChildren({this.node, this.children})
      : assert(node != null || children != null);
}

ScreengraphNode _getRootNode(String rootPath) {
  final rootElement = WidgetsBinding.instance!.renderViewElement!;
  return ScreengraphNode(
    element: rootElement,
    path: rootPath,
  );
}

/// Explore/Geenrate a screengraph for a given [Element]
class ScreengraphExplorer {
  final Map<String, int> _elementPathOccurrences = {};

  /// Explore current screengraph and call the provided [onNodeVisited] on each node visited
  void searchForNodeAtPosition(
      Offset position, Function(ScreengraphNode node) onNodeVisited) {
    _exploreSubtreeAroundPosition(position, _getRootNode(""), onNodeVisited);
  }

  /// Generate a screengraph
  Future<List> generateScreengraph(String rootPath) async {
    return [(await _generateSubScreengraph(_getRootNode(rootPath))).node];
  }

  void _exploreSubtreeAroundPosition(Offset position, ScreengraphNode node,
      Function(ScreengraphNode node) onNodeVisited) {
    node = _updateNodePositionalIndex(node);

    if (!node.frame.contains(position)) {
      return;
    }

    onNodeVisited(node);

    node.visitChildren(
      (ScreengraphNode node) =>
          _exploreSubtreeAroundPosition(position, node, onNodeVisited),
    );
  }

  Future<_ScreengraphNodeOrChildren> _generateSubScreengraph(
    ScreengraphNode node,
  ) async {
    node = _updateNodePositionalIndex(node);

    List<Map> children = [];
    for (DiagnosticsNode childElement
        in node.element.toDiagnosticsNode().getChildren()) {
      _ScreengraphNodeOrChildren child = await _generateSubScreengraph(
        ScreengraphNode(
          element: childElement.value as Element,
          parent: node,
        ),
      );

      if (child.children != null) {
        children = [...children, ...child.children!];
      } else if (child.node != null) {
        children = [...children, child.node!];
      }
    }

    if (isElementPrivate(node.element)) {
      /// if element is private, we do not want to add a node for this corresponding element in the screengraph
      /// we directly return the children
      return _ScreengraphNodeOrChildren(children: children);
    }

    /// otherwise, we create a new node for the current element
    /// and return it's children recursively
    return _ScreengraphNodeOrChildren(
        node: {...(await node.toScreengraphJson()), 'children': children});
  }

  ScreengraphNode _updateNodePositionalIndex(ScreengraphNode node) {
    _updateElementPathOccurrences(node.path);
    final occurence = _elementPathOccurrences[node.path];
    if (occurence != null) {
      /// _elementPathOccurrences[path] is updated by _updateElementPathOccurrences and should not end up null by the time it's called
      node.appendPositionalIndexToPath(occurence);
    }
    return node;
  }

  void _updateElementPathOccurrences(String newPath) {
    final occurence = _elementPathOccurrences[newPath];
    if (occurence != null) {
      _elementPathOccurrences[newPath] = occurence + 1;
    } else {
      _elementPathOccurrences[newPath] = 0;
    }
  }
}
